"""
Support for reading data and metadata from EPREM observer files.
"""

import typing

from ._core import (
    Array,
    Scalar,
    Axis,
    array_factory as array,
    scalar_factory as scalar,
    axis_factory as axis,
)
from ._reference import (
    ARRAYS,
    SCALARS,
    AXES,
)
from ._viewers import (
    HINTS,
    VIEWERS,
    View,
    view_factory as view,
)
from ._interfaces import (
    Axes,
    Arrays,
    Grid,
    axes_factory as axes,
    arrays_factory as arrays,
    grid_factory as grid,
)


Observable = typing.Union[Array, Scalar]
Member = typing.Union[Array, Scalar, Axes]


__all__ = [
    'ARRAYS',
    'SCALARS',
    'AXES',
    'HINTS',
    'VIEWERS',
    'Array',
    'Scalar',
    'Axis',
    'View',
    'Axes',
    'Arrays',
    'Grid',
    'array',
    'scalar',
    'axis',
    'view',
    'axes',
    'arrays',
    'grid',
]





import fractions
import itertools
import numbers
import re
import typing

from .. import etc
from . import _part


class OperandTypeError(TypeError):
    """Type error associated with a symbolic operand."""


class OperandValueError(ValueError):
    """Value error associated with a symbolic operand."""


@etc.autostr
class Operand(_part.Part):
    """An operand in a symbolic expression.

    Algebraic operands mainly exist to support the `~symbolic.Expression` class.
    They may be simple or general, as described below.

    A simple symbolic operand has the form [c]b[^e] or c[^e], where `c` is a
    numerical coefficient, `b` is a string base, and `e` is a numerical
    exponent. Braces ('[]') denote an optional component that defaults to a
    value of 1. The `~symbolic.Term` class formally represents simple symbolic
    operands; the form [c]b[^e] corresponds to a variable term and the form
    c[^e] corresponds to a constant term.

    Examples include:
    
    * `'1'`: unity / multiplicative identity
    * `'1^n'` (`n` real): constant equivalent to unity
    * `'m^n'` (`m`, `n` real): arbitrary constant
    * `'V'`: variable 'V' with coefficient 1 and exponent 1
    * `'V^n'` (`n` real): variable 'V' with coefficient 1 and exponent n
    * `'mV'` (`m` real): variable 'V' with coefficient m and exponent 1
    * `'mV^n'` (`m`, `n` real): variable 'V' with coefficient m and exponent n

    Note that the base of a variable term may comprise multiple characters as
    long as it does not begin with a digit, which this class will interpret as
    part of the coefficient.

    A general symbolic operand consists of simpler operands (though not
    necessarily formally simple operands) combined with symbolic operators and
    separators. All formally simple operands are general operands. The following
    are examples of (non-simple) general symbolic operands:

    * `'a * b^2'` <=> `'(a * b^2)'` <=> `'(a * b^2)^1'`
    * `'(a * b^2)^3'`
    * `'(a * b^2)^3/2'`
    * `'((a / b^2)^3 * c)^2'`
    * `'(a / b^2)^3 * c^2'`
    * `'a / (2 * 4b)'`
    * `'(2a * b)^3 / (4 * c)'`

    There are many more ways to construct a general operand than a simple
    operand. This is by design, to support building instances of
    `~symbolic.Expression` with `~symbolic.Parser`.
    """

    def __init__(
        self,
        coefficient: numbers.Real=None,
        base: str=None,
        exponent: numbers.Real=None,
    ) -> None:
        self.coefficient = fractions.Fraction(coefficient or 1)
        """The numerical coefficient."""
        self.base = base or '1'
        """The base term or complex."""
        self.exponent = fractions.Fraction(exponent or 1)
        """The numerical exponent."""

    @property
    def attrs(self):
        """The current coefficient, base, and exponent."""
        return (self.coefficient, self.base, self.exponent)

    def __pow__(self, power):
        """Create a new operand, raised to `power`."""
        coefficient = self.coefficient ** power
        exponent = self.exponent * power
        return type(self)(coefficient, self.base, exponent)

    def __mul__(self, other):
        """Create a new operand, multiplied by `other`."""
        coefficient = self.coefficient * other
        return type(self)(coefficient, self.base, self.exponent)

    __rmul__ = __mul__

    def __eq__(self, other) -> bool:
        """True if two operands' attributes are equal."""
        if not isinstance(other, Operand):
            return NotImplemented
        if not other.base == self.base:
            return False
        return all(
            float(getattr(other, attr)) == float(getattr(self, attr))
            for attr in ('exponent', 'coefficient')
        )

    def __str__(self) -> str:
        """A simplified representation of this object."""
        return self.format()

    def format(self):
        """Format this operand for printing."""
        if self.coefficient == 1 and self.exponent == 1:
            return self.base
        string = f"({self.base})"
        if self.exponent != 1:
            string = f"{string}^{self.exponent}"
        if self.coefficient != 1:
            string = f"{self.coefficient}{string}"
        return string


class OperandFactory(_part.Factory):
    """A factory that produces symbolic operands."""

    rational = r""" # Modeled after `fractions._RATIONAL_FORMAT`
        [-+]?                 # an optional sign, ...
        (?=\d|\.\d)           # ... only if followed by <digit> or .<digit>
        \d*                   # and a possibly empty numerator
        (?:                   # followed by ...
            (?:/\d+?)         # ... an optional denominator
        |                     # OR
            (?:\.\d*)?        # ... an optional fractional part,
            (?:[eE][-+]?\d+)? #     and an optional exponent
        )
    """
    base = r"""
        [a-zA-Z#_]+ # one or more accepted non-digit character(s)
        \d*         # followed by optional digits
    """

    def __init__(
        self,
        opening: str='(',
        closing: str=')',
        raising: str='^',
    ) -> None:
        exponent = fr'\{raising}{self.rational}'
        self.patterns = {
            'constant': re.compile(
                fr'(?P<coefficient>{self.rational})'
                fr'(?P<exponent>{exponent})?',
                re.VERBOSE,
            ),
            'variable': re.compile(
                fr'(?P<coefficient>{self.rational})?'
                fr'(?P<base>{self.base})'
                fr'(?P<exponent>{exponent})?',
                re.VERBOSE,
            ),
            'complex': re.compile(
                fr'(?P<coefficient>{self.rational})?'
                fr'(?P<base>\{opening}.+?\{closing})'
                fr'(?P<exponent>{exponent})?',
                re.VERBOSE,
            ),
            'exponent': re.compile(exponent, re.VERBOSE),
            'opening': re.compile(fr'\{opening}', re.VERBOSE),
            'closing': re.compile(fr'\{closing}', re.VERBOSE),
            'raising': re.compile(fr'\{raising}', re.VERBOSE)
        }
        """Compiled regular expressions for symbolic operands."""

    _argtypes = {
        'coefficient': numbers.Real,
        'base': str,
        'exponent': (numbers.Real, str),
    }
    @classmethod
    def isvalid(cls, name: str, this: typing.Any):
        """True if `this` is valid for use as the named attribute."""
        return isinstance(this, cls._argtypes[name])

    def normalize(self, *args):
        """Extract attributes from the given argument(s)."""
        try:
            nargs = len(args)
        except TypeError:
            raise OperandTypeError(args) from None
        if nargs == 1:
            return self._length_1(args)
        if nargs == 2:
            return self._length_2(args)
        if nargs == 3:
            return self._length_3(args)
        raise OperandValueError(
            f"{self.__class__.__qualname__}"
            f" accepts 1, 2, or 3 arguments"
            f" (got {nargs})"
        )

    def _length_1(self, args: typing.Any):
        """Normalize a length-1 argument tuple, if possible.

        A length-1 tuple may represent either:

        - coefficient <Real>
        - base <str>

        If it has one of these forms, this method will substitute the default
        value for the missing attributes; otherwise, it will raise an exception.
        """
        arg = args[0]
        names = ('base', 'coefficient')
        for name in names:
            given = {name: arg}
            if self.isvalid(name, arg):
                return self.standardize(fill=True, **given)
        raise OperandTypeError(
            "A single argument may be either"
            " a coefficient <Real> or a base <str>;"
            f" not {type(arg)}"
        )

    def _length_2(self, args: typing.Any):
        """Normalize a length-2 argument tuple, if possible.

        A length-2 tuple may represent either:

        - (base <str>, exponent <Real or str>)
        - (coefficient <Real>, exponent <Real or str>)
        - (coefficient <Real>, base <str>)

        If it has one of these forms, this method will substitute the default
        value for the missing attribute; otherwise, it will raise an exception.
        """
        combinations = itertools.combinations(self._argtypes, 2)
        for names in combinations:
            given = dict(zip(names, args))
            if all(self.isvalid(name, arg) for name, arg in given.items()):
                return self.standardize(fill=True, **given)
        badtypes = [type(arg) for arg in args]
        raise OperandTypeError(
            "Acceptable two-argument forms are"
            " (base <str>, exponent <Real or str>),"
            " (coefficient <Real>, exponent <Real or str>),"
            " or"
            " (coefficient <Real>, base <str>);"
            " not"
            f"({', '.join(str(t) for t in badtypes)})"
        )

    def _length_3(self, args: typing.Any):
        """Normalize a length-3 argument tuple.

        A length-3 tuple must have the form (coefficient <Real>, base <str>,
        exponent <Real or str>).
        """
        return self.standardize(
            coefficient=args[0],
            base=args[1],
            exponent=args[2],
            fill=True,
        )

    def create(self, *args, strict: bool=False) -> typing.Optional[Operand]:
        """Create an operand from input.

        Parameters
        ----------
        *args
            The object(s) from which to create an operand, if possible. This may
            take one of the following forms: a single string representing the
            base operand; a numerical coefficient and a base string; a base
            string and a numerical exponent; or a coefficient, base, and
            exponent. A missing coefficient or exponent will default to 1.

        strict : bool, default=false
            If true, this method will return `None` if it is unable to create an
            operand from `*args`. The default behavior is to return the input
            (with default coefficient and exponent, if necessary) as an instance
            of `~symbolic.Operand`.

        Returns
        -------
        `~symbolic.Operand` or `None`
            An instance of `~symbolic.Operand` or one of its subclasses if the
            input arguments represent a valid operand. The `strict` keyword
            dictates the return behavior when input does not produce an operand.

        Notes
        -----
        This method will create the most general symbolic operand possible from
        the initial string. It will parse a simple symbolic operand into a
        coefficient, variable, and exponent but it will not attempt to fully
        parse a general symbolic operand into simpler operands (i.e. symbolic
        terms). In other words, it will do as little work as possible to extract
        a coefficient and exponent, and the expression on which they operate. If
        all attempts to determine appropriate attributes fail, the value of
        `strict` controls its return behavior.
        
        See note at `parse` for differences between this method and that.

        The following examples use the general symbolic operands described in
        `~symbolic.Operand` to illustrate the minimal parsing described above:

        - `'a * b^2'` <=> `'(a * b^2)'` <=> `'(a * b^2)^1'` -> `1, 'a * b^2', 1`
        - `'2a * b^2'` -> `1, '2a * b^2', 1`
        - `'2(a * b^2)'` -> `2, 'a * b^2', 1`
        - `'(a * b^2)^3'` -> `1, 'a * b^2', 3`
        - `'2(a * b^2)^3'` -> `2, 'a * b^2', 3`
        - `'(a * b^2)^3/2'` -> `1, 'a * b^2', '3/2'`
        - `'((a / b^2)^3 * c)^2'` -> `1, '(a / b^2)^3 * c', 2`
        - `'(a / b^2)^3 * c^2'` -> `1, '(a / b^2)^3 * c^2', 1`
        """
        c0, b0, e0 = self.normalize(*args).values()
        ends = (b0[0], b0[-1])
        if any(self.patterns['raising'].match(c) for c in ends):
            raise OperandValueError(b0) from None
        match = self.search(b0, mode='fullmatch')
        if not match:
            if not strict:
                return Operand(c0, b0, e0)
            return
        if not isinstance(match.result, Operand):
            raise TypeError(
                f"Expected Operand but got {type(match.result)}"
            ) from None
        c1, base, e1 = match.result.attrs
        coefficient = c0 * (c1 ** e0)
        exponent = e1 * e0
        if not isinstance(match.result, Term):
            interior = self.create(base)
            if isinstance(interior, Term):
                coefficient *= interior.coefficient ** exponent
                base = interior.base
                exponent *= interior.exponent
                return term_factory(
                    coefficient=coefficient,
                    base=base,
                    exponent=exponent,
                )
            return Operand(
                coefficient=coefficient,
                base=base,
                exponent=exponent,
            )
        return term_factory(
            coefficient=coefficient,
            base=base,
            exponent=exponent,
        )

    def parse(self, string: str):
        """Extract an operand at the start of `string`, possible.
        
        Notes
        -----
        The primary difference between `~create` and `~parse` is as follows:

        - `~create` resolves input into (coefficient, base, exponent), then
          creates an appropriate operand from the base string, then applies the
          input coefficient and exponent, and finally returns the operand.

        - `~parse` attempts to match an operand at the start of a string, then
          creates an appropriate operand from only that substring, and finally
          returns the operand and the remainder of the string.
        """
        stripped = string.strip()
        if match := self.search(stripped):
            return match

    def search(self, string: str, **kwargs):
        """Search for an operand in the given string.
        
        Parameters
        ----------
        string
            The string to which to apply pattern-matching methods in an attempt
            to find an appropriate operand.

        **kwargs
            Keyword arguments to pass to the pattern-matching methods.

        Returns
        -------
        `~symbolic.PartMatch` or `None`
            An object representing the matched substring and contextual
            information, if any attemp to match was successful. If no method
            found an operand in `string` (subject to any given keyword
            arguments), this method will return `None`.
        """
        methods = (
            self._match_simplex,
            self._match_complex,
        )
        if match := etc.apply(methods, string, **kwargs):
            return match

    def _match_simplex(
        self,
        string: str,
        mode: str='match',
        start: int=0,
    ) -> typing.Optional[_part.Match[Operand]]:
        """Attempt to find an irreducible term at the start of `string`.

        Notes
        -----
        This method tries to match the 'variable' pattern before the 'constant'
        pattern because `re.match` will find a match for 'constant' at the start
        of any variable term with an explicit coefficient.
        """
        target = string[start:]
        matches = {
            key: self._get_match_method(key, mode)(target)
            for key in ('variable', 'constant')
        }
        if not any(matches.values()):
            return
        if all(matches.values()):
            same = matches['variable'][0] == matches['constant'][0]
            key = 'constant' if same else 'variable'
            build_method = self._get_build_method(key)
            return build_method(matches[key])
        for key, match in matches.items():
            if match:
                return self._get_build_method(key)(match)

    def _get_match_method(
        self,
        pattern: str,
        mode: str,
    ) -> typing.Callable[[str], re.Match]:
        """Look up the appropriate matching method for `pattern` and `mode`."""
        return getattr(self.patterns[pattern], mode)

    def _get_build_method(
        self,
        pattern: str,
    ) -> typing.Callable[[re.Match], _part.Match[Operand]]:
        """Look up the appropriate building method for `pattern`."""
        return getattr(self, f'_build_{pattern}')

    def _build_variable(self, match: re.Match):
        """Build a variable term from a match object."""
        standard = self.standardize(**match.groupdict(), fill=True)
        return _part.Match(
            result=term_factory(**standard),
            context=match,
        )

    def _build_constant(self, match: re.Match):
        """Build a constant term from a match object."""
        standard = self.standardize(**match.groupdict(), fill=True)
        coefficient = standard['coefficient'] ** standard['exponent']
        return _part.Match(
            result=term_factory(coefficient=float(coefficient)),
            context=match,
        )

    def _match_complex(
        self,
        string: str,
        mode: str='match',
        start: int=0,
    ) -> typing.Optional[_part.Match[Operand]]:
        """Attempt to match a complex operand at the start of `string`."""
        target = string[start:]
        bounds = self.find_bounds(target)
        if not bounds:
            return
        i0, end = bounds
        result = {'base': target[i0+1:end-1]}
        if match := self._match_simplex(target[:i0], mode='fullmatch'):
            result['coefficient'] = match.result.coefficient
            i0 = 0
        if mode == 'match' and i0 != 0:
            return
        if exp := self.patterns['exponent'].match(target[end:]):
            result['exponent'] = exp[0]
            end += exp.end()
        if mode == 'fullmatch' and (i0, end) != (0, len(target)):
            return
        standard = self.standardize(**result, fill=True)
        return _part.Match(
            result=Operand(**standard),
            context={'end': end, 'start': start, 'string': string},
        )

    def find_bounds(self, string: str):
        """Find the indices of the first bounded substring, if any.
        
        A bounded substring is any portion of `string` that is bounded on the
        left by the opening separator and on the right by the closing separator.
        Opening and closing separators are an immutable attribute of an instance
        of this class.

        Parameters
        ----------
        string
            The string in which to search for a bounded substring.

        Returns
        -------
        tuple of int, or `None`
            The index of the leftmost opening separator and the index of the
            first character beyond the rightmost closing separator (possibly the
            end), if there is a bounded substring; otherwise, `None`. The
            convention is such that if `start, end = find_bounds(string)`,
            `string[start:end]` will produce the bounded substring with bounds.

        Examples
        --------
        Define a list of test strings::

            >>> strings = [
            ...     '(a*b)',
            ...     '(a*b)^2',
            ...     '3(a*b)',
            ...     '3(a*b)^2',
            ...     '3(a*b)^2 * (c*d)',
            ...     '4a^4',
            ...     '3(4a^4)^3',
            ...     '2(3(4a^4)^3)^2',
            ... ]

        Create an instance of this class with the default operators and
        separators::

            >>> operand = symbolic.OperandFactory()

        Find the bounding indices of each test string, if any, and display the
        corresponding substring::

            >>> for string in strings:
            ...     bounds = operand.find_bounds(string)
            ...     print(f"{bounds}: {string!r} -> ", end='')
            ...     if bounds:
            ...         start, end = bounds
            ...         result = f"{string[start:end]}"
            ...     else:
            ...         result = string
            ...     print(f"{result!r}")
            ... 
            (0, 5): '(a*b)' -> '(a*b)'
            (0, 5): '(a*b)^2' -> '(a*b)'
            (1, 6): '3(a*b)' -> '(a*b)'
            (1, 6): '3(a*b)^2' -> '(a*b)'
            (1, 6): '3(a*b)^2 * (c*d)' -> '(a*b)'
            None: '4a^4' -> '4a^4'
            (1, 7): '3(4a^4)^3' -> '(4a^4)'
            (1, 12): '2(3(4a^4)^3)^2' -> '(3(4a^4)^3)'
        """
        initialized = False
        count = 0
        i0 = 0
        for i, c in enumerate(string):
            if self.patterns['opening'].match(c):
                count += 1
                if not initialized:
                    i0 = i
                    initialized = True
            elif self.patterns['closing'].match(c):
                count -= 1
            if initialized and count == 0:
                return i0, i+1

    def standardize(
        self,
        fill: bool=False,
        **given
    ) -> typing.Dict[str, typing.Union[float, int, str, fractions.Fraction]]:
        """Cast to appropriate types and fill in defaults, if necessary."""
        full = {
            'coefficient': {'callable': self._standard_coefficient},
            'base': {'callable': self._standard_base},
            'exponent': {'callable': self._standard_exponent},
        }
        default = self.fill_defaults(**dict.fromkeys(full.keys()))
        updatable = full.copy() if fill else {k: full[k] for k in given}
        return {
            key: attr['callable'](given.get(key) or default[key])
            for key, attr in updatable.items()
        }

    def _standard_coefficient(self, v):
        """Convert input to a standard coefficient."""
        return fractions.Fraction(v or 1)

    def _standard_base(self, v):
        """Convert input to a standard base."""
        return str(v)

    def _standard_exponent(self, v):
        """Convert input to a standard exponent."""
        if isinstance(v, str):
            v = self.patterns['raising'].sub('', v)
        return fractions.Fraction(v or 1)

    def fill_defaults(self, **given):
        """Return the default value for any explicitly null arguments.

        If the given key-value pairs contain an argument with a null value
        (e.g., `None`), this method will replace it with the default value. It
        will pass all other values through unaltered and will not fill in
        default values corresponding to other keys.
        """
        defaults = {
            'coefficient': 1,
            'base': '1',
            'exponent': 1,
        }
        given.update(
            {
                key: defaults[key]
                for key, value in given.items()
                if key in defaults and not value
            }
        )
        return given

    def strip_separators(self, string: str):
        """Remove one opening and one closing separator."""
        opened = self.patterns['opening'].match(string[0])
        closed = self.patterns['closing'].match(string[-1])
        if not (opened and closed):
            return string
        string = self.patterns['opening'].sub('', string, count=1)
        string = self.patterns['closing'].sub('', string[::-1], count=1)
        return string[::-1]


class Term(Operand):
    """A symbolic operand with an irreducible base.
    
    In the single-argument form, the user provides only the base quantity by
    positional-only argument. In the triple-argument form, the user provides the
    coefficient, base, and exponent by positional or keyword argument(s), with
    the caveat that if one argument is positional, they must all be positional.
    This restriction prevents ambiguity among the possible double-argument
    combinations.
    """

    # NOTE: Currently redundant with `OperandFactory`.

    rational = r""" # Modeled after `fractions._RATIONAL_FORMAT`
        [-+]?                 # an optional sign, ...
        (?=\d|\.\d)           # ... only if followed by <digit> or .<digit>
        \d*                   # and a possibly empty numerator
        (?:                   # followed by ...
            (?:/\d+?)         # ... an optional denominator
        |                     # OR
            (?:\.\d*)?        # ... an optional fractional part,
            (?:[eE][-+]?\d+)? #     and an optional exponent
        )
    """
    base = r"""
        [a-zA-Z#_]+ # one or more accepted non-digit character(s)
        \d*         # followed by optional digits
    """
    _base_re = re.compile(fr'({rational}|{base})', re.VERBOSE)

    @classmethod
    def base_is_valid(cls, base):
        """True if `base` can initialize a symbolic term."""
        return bool(cls._base_re.fullmatch(str(base)))

    def __call__(self, value: numbers.Real):
        """Evaluate a variable term at this value.
        
        This method will attempt to substitute `value` for this term's `base`
        attribute. If successful, it will return a constant term that the caller
        may cast to `int` or `float` type.

        Parameters
        ----------
        value : real
            The numerical value at which to evaluate this term.

        Returns
        -------
        `~symbolic.Term`
            A new instance of this class equivalent to the constant numerical
            value of this term when `base == value`.
        """
        if not isinstance(value, numbers.Real):
            errmsg = f"Can't evaluate term with value {value!r}"
            raise TypeError(errmsg)
        if self.base != '1':
            coefficient = self.coefficient * (value ** self.exponent)
            return type(self)(coefficient=coefficient)
        errmsg = f"Can't evaluate term with base {self.base!r}"
        raise TypeError(errmsg)

    def format(self, style: str=None):
        """Format this term.

        Parameters
        ----------
        style : string, optional
            Select a the formatting style. The default behavior is to represent
            this term in the form `'cb^e'`, where `c` is the coefficient, `b` is
            the base, and `e` is the exponent. See Styles for supported styles.

        Styles
        ------
        This section lists the supported formatting styles. All strings are
        case-insensitive unless otherwise noted.
        
        - `'tex'`: format the term for TeX-like display.
        """
        coefficient = self._format_coefficient()
        if self.base == '1':
            return f"{coefficient}"
        fmt = style or ''
        exponent = self._format_exponent(fmt)
        if 'tex' in fmt.lower():
            if self.base == '#':
                return rf"${coefficient}{self.base}{exponent}$"
            return rf"${coefficient}\mathrm{{{self.base}}}{exponent}$"
        return f"{coefficient}{self.base}{exponent}"

    def _format_coefficient(self):
        """Format the coefficient for printing."""
        if self.base != '1' and self.coefficient == 1:
            return ''
        if self._ambiguous_coefficient():
            return f"({self.coefficient})"
        if float(self.coefficient) == int(self.coefficient):
            return str(int(self.coefficient))
        return str(self.coefficient)

    def _ambiguous_coefficient(self):
        """True if this term's coefficient needs to be grouped."""
        return (
            isinstance(self.coefficient, fractions.Fraction)
            and self.base != '1'
            and self.coefficient.denominator != 1
        )

    def _format_exponent(self, style: str):
        """Format the current exponent for printing."""
        if self.base == '1' or self.exponent == 1:
            return ''
        if not style:
            return f"^{self.exponent}"
        if 'tex' in style.lower():
            return f"^{{{self.exponent}}}"
        raise ValueError(f"Can't format {self.exponent}")

    def __int__(self):
        """Called for int(self)."""
        return self._cast_to(int)

    def __float__(self):
        """Called for float(self)."""
        return self._cast_to(float)

    _T = typing.TypeVar('_T', int, float)

    def _cast_to(self, __type: _T) -> _T:
        """Internal method for casting to numeric type."""
        if self.base == '1':
            return __type(self.coefficient)
        errmsg = f"Can't convert term with base {self.base!r} to {__type}"
        raise TypeError(errmsg) from None

    def __eq__(self, other) -> bool:
        if isinstance(other, numbers.Real):
            return float(self) == float(other)
        if isinstance(other, str):
            term = OperandFactory().create(other)
            return super().__eq__(term)
        return super().__eq__(other)

    def __hash__(self) -> int:
        # NOTE: If we decide to implement `Operand.__hash__`, we will still need
        # to explicitly define `Term.__hash__`. See
        # https://docs.python.org/3/reference/datamodel.html#object.__hash__ for
        # an explanation.
        return hash(self.attrs)


@typing.overload
def term_factory(base: str, /) -> Term:
    """Create a symbolic term from a base quantity.
    
    Parameters
    ----------
    base : string
        The base quantity of this symbolic term.
    """

@typing.overload
def term_factory(
    coefficient: numbers.Real=1,
    base: str='1',
    exponent: numbers.Real=1,
) -> Term:
    """Create a symbolic term from coefficient, base, and exponent.
    
    Parameters
    ----------
    coefficient : real number, default=1
        The numerical coefficient to associate with the base quantity.

    base : string, default='1'
        The base quantity of this symbolic term.

    exponent : real number, default=1
        The exponent of the base quantity.
    """

def term_factory(*args, **kwargs):
    """Factory function for symbolic terms."""
    if not kwargs:
        # positional argument(s) only
        if not args:
            # use all default values
            return Term()
        if len(args) == 1:
            # single-argument form (positional base only)
            base = args[0]
            _validate_term_base(base)
            return Term(base=base)
        if len(args) == 3:
            # triple-argument form
            base = args[1]
            _validate_term_base(base)
            return Term(*args)
    if not args:
        # keyword arguments only
        if 'base' in kwargs:
            # NOTE: We don't use `base = kwargs.get('base')` because we
            # don't want to mistakenly treat `None` as the base when
            # comparing to the base RE pattern. We'll let `Operand`
            # determine the default value.
            _validate_term_base(kwargs['base'])
        return Term(**kwargs)
    raise TypeError(
        f"Cannot create a symbolic term from {args} and {kwargs}."
    ) from None


def _validate_term_base(base):
    """Raise an exception if `base` does not match the definition."""
    if not Term.base_is_valid(base):
        raise ValueError(
            f"Cannot create a symbolic term with base {base!r}"
        ) from None


def asterms(these: typing.Iterable[str]):
    """Convert strings to terms, if possible."""
    return [OperandFactory().create(this) for this in these]





"""
Support for defining and working with labeled axes.
"""

import collections.abc
import numbers
import typing

import numpy
import numpy.typing

from .. import container
from .. import etc
from .. import numeric
from .. import quantity
from . import _axis
from ..typehelp import Self


AxisType = typing.Union[
    typing.Iterable[int],
    typing.Iterable[str],
    _axis.Axis,
]


@etc.autostr
class Axes(collections.abc.Mapping, typing.Mapping[str, _axis.Axis]):
    """A mapping from dimension to axis."""
    def __init__(self, mapping: typing.Dict[str, _axis.Axis]) -> None:
        self._mapping = mapping
        self._dimensions = None
        self._names = None

    def __len__(self) -> int:
        """Called for len(self)."""
        return len(self._mapping)

    def __iter__(self) -> typing.Iterator[str]:
        """Called for iter(self)."""
        return iter(self._mapping)

    def __getitem__(self, arg: typing.Union[str, typing.SupportsIndex], /):
        """Called for self[arg]."""
        if isinstance(arg, str):
            # User requested an axis by name.
            return self._mapping[arg]
        try:
            # User requested an axis by index. Note that this relies on
            # order-preserving `dict`. Even though this became a standard
            # language feature in Python 3.7, users should prefer name-based
            # access.
            key = self.names[arg]
        except IndexError as err:
            raise KeyError(f"Unrecognized axis key: {arg}") from err
        return self._mapping[key]

    def __str__(self) -> str:
        """A printable representation of this object."""
        items = ', '.join(f"{k}: {a!r}" for k, a in self._mapping.items())
        return f"{{{items}}}"

    def __repr__(self) -> str:
        """An unambiguous representation of this object."""
        items = ', '.join(f"{k}: {a!r}" for k, a in self._mapping.items())
        return f"{self.__class__.__qualname__}({items})"

    def __eq__(self, other):
        """Called for self == other."""
        if not isinstance(other, Axes):
            return NotImplemented
        if self.names != other.names:
            return False
        for k, this in self.items():
            that = other[k]
            if not numpy.array_equal(this, that):
                return False
        return True

    def __lt__(self, other):
        """Called for self < other."""
        if not isinstance(other, Axes):
            return NotImplemented
        akeys = self.keys()
        bkeys = other.keys()
        ab = akeys & bkeys
        for k in ab:
            if self[k] != other[k]:
                raise ValueError(
                    f"Cannot apply '<' to instances of {type(self)}"
                    f" with different values of {k!r}"
                ) from None
        return akeys < bkeys

    def __le__(self, other):
        """Called for self <= other."""
        if not isinstance(other, Axes):
            return NotImplemented
        akeys = self.keys()
        bkeys = other.keys()
        ab = akeys & bkeys
        for k in ab:
            if self[k] != other[k]:
                raise ValueError(
                    f"Cannot apply '<=' to instances of {type(self)}"
                    f" with different values of {k!r}"
                ) from None
        return akeys <= bkeys

    def __gt__(self, other):
        """Called for self > other."""
        if not isinstance(other, Axes):
            return NotImplemented
        akeys = self.keys()
        bkeys = other.keys()
        ab = akeys & bkeys
        for k in ab:
            if self[k] != other[k]:
                raise ValueError(
                    f"Cannot apply '>' to instances of {type(self)}"
                    f" with different values of {k!r}"
                ) from None
        return akeys > bkeys

    def __ge__(self, other):
        """Called for self >= other."""
        if not isinstance(other, Axes):
            return NotImplemented
        akeys = self.keys()
        bkeys = other.keys()
        ab = akeys & bkeys
        for k in ab:
            if self[k] != other[k]:
                raise ValueError(
                    f"Cannot apply '>=' to instances of {type(self)}"
                    f" with different values of {k!r}"
                ) from None
        return akeys >= bkeys

    def __add__(self, other):
        """Called for self + other.

        This method will fill singular dimensions in each instance with the
        corresponding full axis in the other instance, if possible.
        """
        if not isinstance(other, Axes):
            return NotImplemented
        if self.keys() != other.keys():
            raise ValueError(
                f"Cannot fill {a} from {b} with different dimensions"
            ) from None
        new = {}
        for key in self:
            a = self[key]
            b = other[key]
            if a == b:
                new[key] = a
            elif len(b) == 1 and b in a:
                new[key] = a
            elif len(a) == 1 and a in b:
                new[key] = b
            else:
                raise ValueError(
                    f"The axes {a} and {b} have inconsistent values"
                ) from None
        return Axes(new)

    def __or__(self, other):
        """Called for self | other."""
        if isinstance(other, typing.Mapping):
            return self._merge(other)
        return NotImplemented

    def _merge(self, other: typing.Mapping):
        """Merge self with other while preserving order."""
        these = list(self)
        those = list(other)
        try:
            keys = container.merge(these, those)
        except container.MergeError as err:
            errmsg = str(err).replace('entries', 'dimensions')
            raise ValueError(errmsg) from err
        mapping = {k: other[k] if k in other else self[k] for k in keys}
        return Axes(mapping)

    def copy(self):
        """Create a shallow copy of this collection."""
        return Axes({d: self[d] for d in self.dimensions})

    @typing.overload
    def replace(self, dimension: str, axis: _axis.Axis) -> Self: ...

    @typing.overload
    def replace(self, dimension: str, **pair: _axis.Axis) -> Self: ...

    def replace(self, *args, **pair):
        """Replace the axis object corresponding to `dimension`.

        Parameters
        ----------
        dimension : string
            The current name of the axis to replace. The given dimension must
            exist in this collection.

        axis : `~Axis`
            The new axis object that will correspond to `dimension`. May not be
            specified along with a key-value pair.

        **pair : `~Axis`
            A key-value pair in which the value is an axis object and the key is
            its corresponding name. This name will replace `dimension` in the
            result. May not be specified along with a positional axis.

        Returns
        -------
        `~Axes`
            A new instance with all the elements in this instance not
            corresponding to `dimension`, and a new dimension-axis pair
            determined from arguments.

        Raises
        ------
        `KeyError`
            The user passed a non-existent dimension.
        `ValueError`
            The user passed a positional axis and a key-value pair.
        """
        dimension, axis = args if len(args) == 2 else (args[0], None)
        if dimension not in self:
            raise KeyError(
                f"Axes do not contain {dimension!r}"
            ) from None
        if pair and axis is not None:
            strargs = etc.strargs(axis, **pair)
            raise TypeError(
                f"Got conflicting axis updates: {strargs}"
            ) from None
        if axis is not None:
            mapping = {
                d: axis if d == dimension else self[d]
                for d in self.dimensions
            }
            return Axes(mapping)
        if pair:
            if len(pair) > 1:
                raise ValueError(
                    "Can only replace one axis at a time."
                ) from None
            key, value = pair.popitem()
            mapping = {}
            for d in self.dimensions:
                if d == dimension:
                    mapping[key] = value
                else:
                    mapping[d] = self[d]
            return Axes(mapping)
        raise ValueError("Missing axis") from None

    @typing.overload
    def insert(
        self,
        dimension: str,
        new: _axis.Axis,
        index: typing.Optional[int]=None,
    ) -> Self: ...

    @typing.overload
    def insert(
        self,
        dimension: str,
        new: _axis.Axis,
        before: typing.Optional[str]=None,
    ) -> Self: ...

    @typing.overload
    def insert(
        self,
        dimension: str,
        new: _axis.Axis,
        after: typing.Optional[str]=None,
    ) -> Self: ...

    def insert(self, dimension, new, **kwargs):
        """Insert dimension-axis pairs."""
        if len(kwargs) > 1:
            strkws = etc.strargs(**kwargs)
            raise TypeError(
                f"Got multiple location keywords for new axes: {strkws}"
            ) from None
        if len(kwargs) == 0:
            return self | {dimension: new}
        locator, location = kwargs.popitem()
        mapping = {}
        for i, d in enumerate(self.dimensions):
            current = self[d]
            if locator == 'before' and d == location:
                mapping[d] = current
                mapping[dimension] = new
            elif locator == 'after' and d == location:
                mapping[dimension] = new
                mapping[d] = current
            elif locator == 'index' and i == location:
                mapping[dimension] = new
                mapping[d] = current
            else:
                mapping[d] = current
        return Axes(mapping)

    def without(self, dimension: str, strict: bool=False):
        """Create a set of axes without the named element.
        
        Parameters
        ----------
        dimension : string
            The name of the axis to remove.

        strict : bool, default=False
            If true and these axes do not contain an element corresponding to
            `dimension`, raise an exception. The default behavior is to ignore a
            missing dimension.

        Returns
        -------
        `~Axes`
            A new instance with all the elements in this instance, except for
            that corresponding to `dimension`.

        Raises
        ------
        `KeyError`
            The user passed a non-existent dimension with `strict = True`.
        """
        if strict and dimension not in self:
            raise KeyError(
                f"Axes does not contain {dimension!r}"
            ) from None
        mapping = {k: v for k, v in self.items() if k != dimension}
        return Axes(mapping)

    def extract(self, *dimensions: str):
        """Extract a subset of these axes.
        
        Parameters
        ----------
        *dimensions
            One or more axis names from which to create the new instance.

        Returns
        -------
        `~Axes`
            A new instance with elements from this instance corresponding to the
            given dimensions.

        Raises
        ------
        `~KeyError`
            These axes do not contain a given dimension.
        `~TypeError`
            The user did not provide any dimensions.
        """
        if not dimensions:
            raise TypeError(
                "No dimensions to extract"
            ) from None
        return Axes({d: self[d] for d in dimensions})

    @typing.overload
    def permute(
        self,
        order: typing.Iterable[typing.SupportsIndex],
    ) -> Self: ...

    @typing.overload
    def permute(
        self,
        *order: typing.SupportsIndex,
    ) -> Self: ...

    def permute(self, *args, **kwargs):
        """Reorder these axes."""
        if args and 'order' in kwargs:
            raise TypeError("Too many arguments") from None
        if args:
            if len(args) == 1:
                return self._permute(args[0])
            return self._permute(args)
        if 'order' in kwargs:
            return self._permute(kwargs['order'])
        raise TypeError("No arguments") from None

    def _permute(self, args: typing.Iterable):
        """Helper for `~permute`."""
        remapped = {d: self[d] for d in self._permute_dimensions(args)}
        return axes_factory(remapped)

    def _permute_dimensions(self, args: typing.Iterable):
        """Helper for `~permute`."""
        if etc.allisinstance(typing.SupportsIndex, *args):
            return self.dimensions.permute(args)
        if etc.allisinstance(str, *args):
            self.dimensions.mismatch(args)
            return numeric.dimensions(args)
        raise TypeError(*args)

    @property
    def dimensions(self):
        """The equivalent `~Dimensions` instance."""
        if self._dimensions is None:
            self._dimensions = numeric.Dimensions(self.names)
        return self._dimensions

    @property
    def names(self) -> typing.Tuple[str]:
        """The axis names."""
        if self._names is None:
            self._names = tuple(self._mapping.keys())
        return self._names


def axes_factory(*args, **kwargs) -> Axes:
    """Internal factory for `~Axes` instances."""
    return Axes(_map_axes(*args, **kwargs))


def _map_axes(*args, **kwargs):
    """Create a `dict` of axes to initialize `~Axes`."""
    if not args:
        # Strip known keyword arguments from the full collection of keyword
        # arguments, leaving only dimension-axis pairs (if any).
        shape = kwargs.pop('shape', None)
        dimensions = kwargs.pop('dimensions', None)
        axes = kwargs.pop('axes', None)
        mapping = kwargs.pop('mapping', None)
        # Merge dimension-axis mappings.
        pairs = kwargs.copy()
        pairs.update(mapping or {})
        if pairs:
            # Create axes from merged dimension-axis pairs.
            return {k: convert_to_axis(arg) for k, arg in pairs.items()}
        if axes is not None:
            if dimensions is None:
                # Create axes from default dimensions and the given axes.
                return {
                    f'x{i}': convert_to_axis(axis)
                    for i, axis in enumerate(axes)
                }
            # Make sure number of dimensions == number of axes.
            na = len(axes)
            nd = len(dimensions)
            if na != nd:
                raise ValueError(
                    f"Number of dimensions ({nd}) != number of axes ({na})"
                ) from None
            # Create axes from the given dimensions and axes.
            return {
                d: convert_to_axis(axis) for d, axis in zip(dimensions, axes)
            }
        if shape is not None:
            if dimensions is None:
                # Create axes from default dimensions and the given shape.
                return {
                    f'x{i}': _axis.Points(range(n))
                    for i, n in enumerate(shape)
                }
            # Make sure number of dimensions == length of shape.
            ns = len(shape)
            nd = len(dimensions)
            if ns != nd:
                raise ValueError(
                    f"Number of dimensions ({nd}) != length of shape ({ns})"
                ) from None
            # Create axes from the given shape and dimensions.
            return {
                k: _axis.Points(range(n)) for k, n in zip(dimensions, shape)
            }
        if dimensions is not None:
            # We're here because we can't create axes from dimensions without
            # knowledge of at least the length of each axis.
            raise TypeError(
                "Cannot create axes from dimensions without shape or axes"
            ) from None
        # We're here because there are no positional arguments, there are no
        # explicit dimension-axis pairs, and all the keyword arguments are
        # `None`, meaning we have nothing from which to create axes.
        raise ValueError("Empty axes") from None
    if len(args) == 1:
        # This ignores all other arguments on purpose.
        arg = args[0]
        if isinstance(arg, typing.Mapping):
            # Create axes from the given mapping.
            return {k: convert_to_axis(a) for k, a in arg.items()}
        raise TypeError(
            f"Expected positional argument to be a mapping, not {type(arg)}"
        ) from None
    if not kwargs and etc.allisinstance(numbers.Integral, *args):
        # Create axes from one or more axis lengths.
        return {f'x{i}': _axis.Points(range(n)) for i, n in enumerate(args)}
    raise TypeError(
        f"Cannot create axes from {etc.strargs(*args, **kwargs)}"
    ) from None


def convert_to_axis(arg):
    """Attempt to convert `arg` to a valid axis object."""
    try:
        axis = _convert_to_axis(arg)
    except (TypeError, quantity.ParsingTypeError):
        raise ValueError(
            f"Cannot convert {arg} to an axis."
        ) from None
    return axis


def _convert_to_axis(arg):
    """Return a valid axis object."""
    if isinstance(arg, _axis.Axis):
        return arg
    if all(isinstance(x, str) for x in arg):
        return _axis.Symbols(arg)
    if all(isinstance(x, numbers.Integral) for x in arg):
        return _axis.Points(arg)
    m = quantity.measure(arg)
    if m.unit == '1' and all(isinstance(x, numbers.Integral) for x in m):
        return _axis.Points(m)
    return _axis.Coordinates(m)





"""
Support for working with EPREM runtime configuration files.
"""

import collections.abc
import contextlib
import functools
import numbers
import pathlib
import re
import typing

from .. import etc
from .. import paths
from .. import symbolic
from . import _src


class Defaults(collections.abc.Mapping):
    """Default parameter values."""

    def __init__(self, srcdir: paths.PathLike) -> None:
        self._basetypes = _src.BaseTypesH(srcdir=srcdir)
        self._reference = _src.ConfigurationC(srcdir=srcdir)
        self._mapping = {
            key: _convert(parameter['defaultVal'], parameter['type'])
            for key, parameter in self.reference.items()
        }

    def __iter__(self) -> typing.Iterator[str]:
        """Called for iter(self)."""
        return iter(self.reference)

    def __len__(self) -> int:
        """Called for len(self)."""
        return len(self.reference)

    def __getitem__(self, name: str):
        """Retrieve a parameter value by name."""
        if name in self.reference:
            return self._evaluate(name)
        raise KeyError(f"Unknown parameter {name!r}")

    _substitutions = {
        'N_PROCS': None,
        'third': 1/3,
    }

    def _evaluate(self, current: typing.Union[str, numbers.Real]):
        """Compute the final value of a parameter."""
        if current == 'abundance': # special case
            return [1.0] * self._evaluate('config.numSpecies')
        if isinstance(current, numbers.Real):
            return current
        if value := cast(current, strict=False):
            return value
        if current in self._substitutions:
            return self._substitutions[current]
        argument = self._get_argument(current)
        if argument is not None:
            definition = argument['definition']
            if isinstance(definition, argument['type']):
                return definition
            return self._evaluate(definition)
        if isinstance(current, str):
            return self._resolve(current)
        raise TypeError(f"Can't evaluate {current!r}") from None

    _struct_member = re.compile(r'\Aconfig\.\w*\Z')

    def _resolve(self, definition: str):
        """Resolve a parameter definition into simpler components."""
        if self._struct_member.match(definition):
            return self._evaluate(definition.replace('config.', ''))
        interior = definition.strip('[]')
        if self._struct_member.match(interior):
            return [self._evaluate(interior.replace('config.', ''))]
        if result := self._compute_sum(definition):
            return result
        if any(c in definition for c in {'*', '/'}):
            expression = symbolic.expression(definition.replace('config.', ''))
            evaluated = [
                term.coefficient * self._evaluate(term.base)**term.exponent
                for term in expression
            ]
            return functools.reduce(lambda x,y: x*y, evaluated)
        raise TypeError(f"Can't resolve {definition!r}") from None

    def _compute_sum(self, arg: str) -> numbers.Real:
        """Compute the sum of two known parameters.
        
        Notes
        -----
        This is only designed to handle strings that contain a single additive
        operator joining two arguments that `_evaluate` already knows how to
        handle.
        """
        for operator in ('+', '-'):
            if operator in arg:
                terms = [
                    self._evaluate(s.strip())
                    for s in arg.split(operator)
                ]
                return terms[0] + float(f'{operator}1')*terms[1]

    def _get_argument(self, name: str):
        """Get the current definition and type of a parameter."""
        if name in self._mapping:
            return {
                'definition': self._mapping[name],
                'type': self.reference[name]['type']
            }
        if name in self.basetypes:
            return {
                'definition': self.basetypes[name],
                'type': type(self.basetypes[name]),
            }

    @property
    def reference(self):
        """Parameter metadata from `configuration.c`."""
        return self._reference

    @property
    def basetypes(self):
        """Values of constants defined in `baseTypes.h`."""
        return self._basetypes


def defaults_factory(srcdir: typing.Optional[paths.PathLike]=None):
    """Factory function for collections of default parameter values."""
    return Defaults(srcdir)


class ConfigKeyError(KeyError):
    pass


class ConfigDiff(typing.NamedTuple):
    """Return type of `ConfigFile.diff`."""

    keys: typing.Dict[int, set]
    full: typing.Dict[int, dict]


@etc.autostr
class ConfigFile(collections.abc.Mapping):
    """A class to handle EPREM run configuration files.

    Notes
    -----
    This interface was designed to provide a faithful representation of the
    information in a given configuration file. A few notable consequences are as
    follows:

    - The look-up methods do not accept aliases as keys.
    - The parsing routine does not attempt to cast parsed values from strings to
      their underlying types.
    - Values have no knowledge of their associated unit.

    See Also
    --------
    `~parameter.interface`
        Create an instance of `~parameter.Interface`, which supports aliased
        look-up, converts values to their underlying types, and returns a
        real-valued object with an associated unit when applicable.
    """
    def __init__(
        self,
        filepath: pathlib.Path,
        comments: typing.List[str],
    ) -> None:
        """Initialize this instance.

        This class is not intended for direct instantiation and initialization
        by general users. Instead, please see `~parameter.config`.
        """
        self._filepath = filepath
        self._comments = comments
        self._parsed = None

    def __str__(self) -> str:
        """A simplified representation of this object."""
        return str(self.source)

    def __iter__(self) -> typing.Iterable[str]:
        """Called for iter(self)."""
        return iter(self.parsed)

    def __len__(self) -> int:
        """Called for len(self)."""
        return len(self.parsed)

    def __getitem__(self, key: str):
        """Get a value and unit for a configuration option."""
        if key in self.parsed:
            return self.parsed[key]
        raise ConfigKeyError(key)

    def compare(self, __o, **kwargs):
        """Compare this config file to another."""
        diff = self.diff(__o, **kwargs)
        df0 = diff.full[0]
        df1 = diff.full[1]
        keys = sorted(set(df0) | set(df1))
        return {k: (df0.get(k), df1.get(k)) for k in keys}

    def diff(self, __o, **kwargs):
        """Compute the symmetric difference between two config files."""
        def normalize(x):
            if isinstance(x, ConfigFile):
                return x
            if isinstance(x, (str, pathlib.Path)):
                return ConfigFile(x, **kwargs)
            raise TypeError(
                "Argument must be a path-like object"
                f" or an instance of this class (got {type(x)})"
            ) from None
        other = normalize(__o)
        keys = {
            0: set(self) - set(other),
            1: set(other) - set(self),
        }
        full = {
            0: dict(set(self.items())),
            1: dict(set(other.items())),
        }
        return ConfigDiff(keys=keys, full=full)

    @property
    def parsed(self):
        """The parsed parameter-value pairs."""
        if self._parsed is None:
            comments = set(self._comments or ['#'])
            self._parsed = _parse_config(self.source, comments)
        return self._parsed

    @property
    def source(self):
        """The path to the parsed configuration file."""
        return self._filepath


def _parse_config(
    filepath: pathlib.Path,
    comments: typing.Iterable[str]=None,
) -> typing.Dict[str, str]:
    """Parse an EPREM config file into a dictionary.

    This method opens the file that the given filepath points to and reads it
    line-by-line. It ignores lines that begin with a valid comment character and
    parses the rest into key-value pairs. It will automatically strip out inline
    comments.
    """
    pairs = {}
    cmnt = comments or ()
    with filepath.open('r') as fp:
        for line in fp:
            line = line.rstrip('\n')
            if line == '' or line == '\n' or line[0] in cmnt:
                continue
            tmp = paths.strip_inline_comments(line, cmnt)
            key, value = tmp.split('=')
            pairs[key.strip()] = value.strip()
    return pairs


def configfile_factory(
    filepath: paths.PathLike,
    comments: typing.Optional[typing.Iterable[str]]=None,
) -> ConfigFile:
    """Create an interface to an EPREM runtime configuration file.

    Parameters
    ----------
    filepath : string or path
        The path to the simulation configuration file to read. This function
        will convert the argument to a fully-qualified read-only path.

    comments : list of strings, default='#'
        List of single-character strings to interpret as comment markers in the
        configuration file. This function will automatically ignore lines that
        begin with either the empty string or the newline character.

    Returns
    -------
    `~ConfigFile`
        An interactive representation of the EPREM configuration file located at
        the given path.

    Notes
    -----
    - The argument to `filepath` may be relative but it must include the file
      name. For example, suppose the relevant configuration file is called
      `config.ini`. Then, passing `filepath='config.ini'` or
      `filepath='./config.ini'` within the containing directory, as well as
      `filepath='</path/to/dir>/config.ini'` elsewhere (even if `/path/to/dir`
      is relative), would be valid, but passing `filepath='</path/to/dir>'`
      (even if `/path/to/dir` is absolute) would not be valid.
    """
    return ConfigFile(
        paths.fullpath(filepath, strict=True),
        comments=list(comments or ['#']),
    )


_RT = typing.TypeVar('_RT')


def _convert(arg: typing.Union[str, _RT], realtype: _RT):
    """Convert `arg` to its real type, if necessary and possible."""
    if isinstance(arg, realtype):
        return arg
    if realtype == bool:
        return soft_convert(arg, boollike)
    if realtype in {int, float}:
        return soft_convert(arg, realtype)
    if realtype == list:
        return soft_convert(arg, stringlist)
    return arg


def soft_convert(
    string: str,
    convert: typing.Union[_RT, typing.Callable[[str], _RT]],
    acceptable: typing.Union[Exception, typing.Tuple[Exception]]=Exception,
) -> typing.Union[str, _RT]:
    """Convert a string to a different type, if possible.

    This function will use the given callable to attempt to convert the given
    string to a different type. If the conversion fails, this function will
    return the unconverted string.

    Parameters
    ----------
    string
        The string to convert.

    convert : callable
        A callable object that accepts a string argument and returns the
        converted value.

    acceptable : exception or iterable of exceptions
        One or more types of valid exception classes that this function should
        ignore when attempting to convert `string`. The default behavior is to
        accept any subclass of `Exception`. Passing arguments to this keyword
        therefore allows the caller to limit the scope of ignorable errors.

    Returns
    -------
    converted type or string
        The converted value of the given string, if conversion succeeded, or the
        unconverted string.
    """
    try:
        value = convert(string)
    except acceptable or Exception:
        return string
    return value


def boollike(arg: str | typing.SupportsInt) -> typing.Literal[True, False]:
    """Convert a string representation of a boolean to its truth value."""
    if isinstance(arg, str):
        if arg.lower() in {'true', 't'}:
            return True
        if arg.lower() in {'false', 'f'}:
            return False
    return bool(int(arg))


def stringlist(string: str) -> list:
    """Convert a string representation of a list to a list"""
    if not (string.startswith('[') and string.endswith(']')):
        raise TypeError(f"Can't convert {string!r} to a list") from None
    inside = string.strip('[]')
    if not inside:
        return []
    items = [item.strip(" ''") for item in inside.split(',')]
    return [cast(i) for i in items]


_NT = typing.TypeVar('_NT', bound=numbers.Number)


def cast(
    arg: _NT,
    strict: bool=True,
) -> typing.Union[int, float, complex, _NT]:
    """Attempt to convert `arg` to an appropriate numeric type.
    
    Parameters
    ----------
    arg
        The object to convert. May be of any type. If it has a numeric type,
        this function will immediately return it.

    strict : bool, default=True
        If true (default), this function will raise an exception if it can't
        convert `arg` to a numerical type. If false, this function will silently
        return upon failure to convert.

    Returns
    -------
    number
        The numerical value of `arg`, if possible. See description of `strict`
        for behavior after failed conversion.
    """
    if isinstance(arg, numbers.Number):
        return arg
    types = (
        int,
        float,
        complex,
    )
    for t in types:
        with contextlib.suppress(ValueError, TypeError):
            return t(arg)
    if strict:
        raise TypeError(f"Can't convert {arg!r} to a number") from None





import builtins
import contextlib
import numbers
import operator as standard
import typing

import numpy
import numpy.typing
from scipy.interpolate import interp1d

from .. import aliased
from .. import container
from .. import datafile
from .. import etc
from .. import measured
from .. import metric
from .. import numeric
from .. import parameter
from .. import physical
from .. import quantity
from .. import reference
from ..reference import ARRAYS
from ..typehelp import Self


class Interpolant(typing.NamedTuple):
    """Attributes necessary for interpolation."""
    reference: physical.Array
    targets: numpy.ndarray


class Array(physical.Array):
    """A physical array with interpolation support."""

    def interpolate(
        self: Self,
        interpolants: typing.Mapping[str, Interpolant],
    ) -> Self:
        """Interpolate this array to target values."""
        return interpolate(self, interpolants)


def array_factory(x, **kwargs):
    """Create an observable array."""
    if isinstance(x, physical.Array):
        return Array(x.data, x.context, x.axes)
    return physical.array(x, **kwargs)


def interpolate(
    array: physical.Array,
    interpolants: typing.Mapping[str, Interpolant],
) -> Array:
    """Interpolate a physical array.

    Parameters
    ----------
    array : `~Array`
        The array object to interpolate.

    interpolants : mapping from string to `~Interpolant`
    """
    if not interpolants:
        # If there are no interpolants, return the array as-is. This is simply a
        # convenience so that calling code need not check whether `interpolants`
        # is empty before calling this function.
        return array
    workspace = None
    axes = array.axes.copy()
    for dimension, interpolant in interpolants.items():
        workspace = _update_interpolation(
            array,
            interpolant.reference,
            interpolant.targets,
            workspace=workspace,
        )
        s = measured.sequence(interpolant.targets, interpolant.reference.unit)
        if dimension in ARRAYS.aliases['radius']:
            axes = axes.replace('shell', radius=physical.Coordinates(s))
        else:
            axes = axes.replace(dimension, physical.Coordinates(s))
    return array.spawn(workspace, unit=array.unit, axes=axes)


def _update_interpolation(
    array: Array,
    reference: numeric.Array,
    targets: numpy.ndarray,
    workspace: numpy.ndarray=None,
) -> numpy.ndarray:
    """Helper for `~interpolate`."""
    x = numpy.array(array) if workspace is None else workspace
    indices = (array.dimensions.index(d) for d in reference.dimensions)
    dst, src = zip(*enumerate(indices))
    reordered = numpy.moveaxis(x, src, dst)
    interpolated = _apply_interpolation(reordered, reference, targets)
    return numpy.moveaxis(interpolated, dst, src)


def _apply_interpolation(
    data: numpy.typing.ArrayLike,
    reference: numpy.typing.ArrayLike,
    targets: typing.Iterable[float],
) -> numpy.ndarray:
    """Interpolate `data` to target values with respect to `reference`.

    This function will always interpolate over the leading dimension.

    Parameters
    ----------
    data : array-like
        The array-like object to interpolate. This function will convert `data`
        to a `numpy.ndarray`.

    reference : array-like
        The values at which `data` is known. This function will convert
        `reference` to a `numpy.ndarray`.

    targets : iterable of float
        The value(s) to which to interpolate `data`.

    Returns
    -------
    `numpy.ndarray`
        An array with the same shape as `array`, except for the interpolated
        dimension, which will have the same length as the number of values
        `targets`.
    """
    darray = numpy.array(data)
    rarray = numpy.array(reference)
    interpolated = [
        _apply_interp1d(darray, rarray, target)
        for target in targets
    ]
    if rarray.ndim == 2:
        return numpy.swapaxes(interpolated, 0, 1)
    return numpy.array(interpolated)


def _apply_interp1d(
    array: numpy.ndarray,
    reference: numpy.ndarray,
    target: float,
) -> typing.List[float]:
    """Interpolate data to `target` along the leading axis."""
    if target in reference:
        idx = container.nearest(reference, target).index
        return array[idx]
    ndim = reference.ndim
    if ndim == 2:
        interps = [interp1d(x, y, axis=0) for x, y in zip(reference, array)]
        return numpy.array([interp(target) for interp in interps])
    if ndim != 1:
        raise physical._array.NDimError(
            f"The reference array may have 1 or 2 (not {ndim}) dimensions"
        ) from None
    if reference.size == 1:
        # This works because we are interpolating over the leading dimension by
        # definition. Squeezing does not necessarily work because there may be
        # other singular dimensions.
        return array[0]
    interp = interp1d(reference, array, axis=0)
    return interp(target)


class Arguments(aliased.Mapping[str, measured.Object]):
    """Array indices and physical assumptions."""

    def __init__(
        self,
        indices: typing.Dict[str, numeric.index.Sequence],
        **assumptions: measured.Object
    ) -> None:
        aliases = reference.ALIASES | parameter.ALIASES
        super().__init__(assumptions, aliases=aliases)
        self._time = indices.get('time')
        self._shell = indices.get('shell')
        self._species = indices.get('species')
        self._energy = indices.get('energy')
        self._mu = indices.get('mu')
        self._extra = assumptions
        self._indices = None
        self._assumptions = None

    def axes(
        self,
        optional: typing.Iterable[str]=None,
        **extra
    ) -> typing.Dict[str, physical.AxisType]:
        """Build a mapping of indices and pseudo indices.
        
        Parameters
        ----------
        optional : iterable of string, optional
            Zero or more names of arguments to include if they are available.

        **extra
            Zero or more key-value pairs in which the key is an argument name
            and the value is the default value to include if the argument is not
            available.

        Notes
        -----
        - In some cases, the mere presence of a parameter may trigger a certain
          algorithmic path (e.g., interpolation) whereas some cases may require
          a particular argument, then proceed depending on the value. Requesting
          an argument via `optional` is more appropriate for the former while
          requesting an argument via `**extra` is more appropriate for the
          latter.
        """
        indices = self.indices.copy()
        for key in optional or ():
            with contextlib.suppress(KeyError):
                indices[key] = self[key]
        for key, default in extra.items():
            indices[key] = self.get(key, default)
        return indices

    @property
    def assumptions(self):
        """All non-index attributes."""
        if self._assumptions is None:
            assumptions = {
                k: v
                if isinstance(v, (physical.Scalar, physical.Tensor, slice))
                else parameter.variable(quantity.measure(v))
                for k, v in self._extra.items()
            }
            aliases = reference.ALIASES | parameter.ALIASES
            self._assumptions = aliased.Mapping(assumptions, aliases=aliases)
        return self._assumptions

    @property
    def indices(self):
        """All axis indices as a mapping."""
        if self._indices is None:
            self._indices = {
                'time': self.time,
                'shell': self.shell,
                'species': self.species,
                'energy': self.energy,
                'mu': self.mu,
            }
        return self._indices

    @property
    def time(self):
        """The time-dimension index."""
        return self._time

    @property
    def shell(self):
        """The shell-dimension index."""
        return self._shell

    @property
    def species(self):
        """The species-dimension index."""
        return self._species

    @property
    def energy(self):
        """The energy-dimension index."""
        return self._energy

    @property
    def mu(self):
        """The mu-dimension index."""
        return self._mu

    def __str__(self) -> str:
        parts = [f"{k}: {getattr(self, k)}" for k in datafile.AXES]
        if s := super().__str__():
            parts.append(s)
        return ', '.join(parts)


class ArgumentError(Exception):
    """Unknown or invalid argument."""


class Context:
    """A general observing context."""

    def __init__(
        self,
        operator: typing.Callable[[Arguments], physical.Array],
        axes: datafile.Axes,
        grid: datafile.Grid,
    ) -> None:
        """Initialize this context."""
        self._operator = operator
        self._axes = axes
        self._grid = grid

    def __eq__(self, other) -> bool:
        """Called for self == other."""
        return (
            isinstance(other, Context)
            and
            self.axes == other.axes
            and
            self.operator == other.operator
        )

    def apply(self, arg: Arguments) -> physical.Array:
        """Apply the observing operation to existing arguments."""
        return self.operator(arg)

    def implement(self, **user) -> physical.Array:
        """Implement the observing operation within this context."""
        # NOTE: This method should assume for now that `user` contains all the
        # necessary axis indices for the target quantity. That is a reasonable
        # assumption whenever `Operation.__getitem__` calls this method, which
        # is currently the only use case.
        indexed = {}
        parsed = {}
        for k, v in user.items():
            try:
                p = self._parse_user_arg(k, v)
            except physical.AxisValueError:
                # NOTE: We are here because `k` is the name of an axis but `v`
                # is not a value in the corresponding axis array. We will
                # therefore indicate to downstream code that it needs to
                # interpolate the corresponding axis to `v` by storing `v` as an
                # assumption and storing a full slice as the relevant index.
                indexed[k] = self._parse_user_arg(k, slice(None))
                parsed[k] = self._parse_user_arg('', v)
            else:
                if k in self.axes:
                    indexed[k] = p
                else:
                    parsed[k] = p
        for key in {'time', 'energy', 'mu'}:
            if key in parsed:
                x = self.axes[key]
                v = parsed[key].withunit(x.unit)
                w = (float(i) for i in v)
                idx = numeric.index.restrict(x, *w, axis=0)
                indexed[key] = numeric.index.sequence(idx)
        if 'radius' in parsed:
            x = self.grid.r
            rv = parsed['radius'].withunit(x.unit)
            rs = (float(ri) for ri in rv)
            idx = numeric.index.restrict(x, *rs, axis=1)
            indexed['shell'] = numeric.index.sequence(idx)
        return self.operator(Arguments(indices=indexed, **parsed))

    def _parse_user_arg(self, k: str, v):
        """Convert a given argument into the appropriate object."""
        if quantity.isindexlike(v):
            return numeric.index.sequence(v)
        name = reference.AXES.names.get(k, k)
        if name in self.axes:
            axis = self.axes[name]
            if isinstance(v, slice):
                return axis.indices[v]
            return axis.index(v)
        # At this point, we can be reasonably confident that `v` is not an axis
        # index, so we will treat it as a physical assumption.
        if isinstance(v, slice):
            return v
        with contextlib.suppress(quantity.MeasuringTypeError):
            return quantity.measure(v)
        raise ArgumentError(
            f"Argument {k!r} is either unknown or has invalid type"
        ) from None

    @property
    def operator(self):
        """The observing function."""
        return self._operator

    @property
    def axes(self):
        """The array-indexing objects."""
        return self._axes

    @property
    def grid(self):
        """The coordinate arrays."""
        return self._grid


class ObservationError(Exception):
    """There is an error in the observed array."""


class Quantity:
    """An arbitrary observable quantity with an array-like interface."""

    def __init__(
        self,
        context: Context,
        unit: metric.Unit,
        dimensions: numeric.Dimensions,
    ) -> None:
        self._context = context
        self._unit = unit
        self._dimensions = dimensions
        self._ndim = None

    def __repr__(self) -> str:
        """A simplified representation of this object."""
        _str = f"unit={str(self.unit)!r}, dimensions={self.dimensions}"
        return f"{self.__class__.__qualname__}({_str})"

    def __array__(self, *args, **kwargs):
        """Called for conversion to a `numpy.ndarray`.

        Notes
        -----
        - All arguments pass directly to the `numpy` array constructor.
        - This will create an array from the entire observable quantity. If you
          want to create an array from a subset of the observable data, you must
          first subscript this quantity, then call `numpy.array` on the result.
          In fact, this method is simply a convenient shorthand for that
          processes, in the case when the subscription returns the full array.
        """
        return numpy.array(self[:], *args, **kwargs)

    def __getitem__(self, args):
        """Access observable data via array subscription."""
        # If `args` is an instance of `Arguments`, we are computing an
        # intermediate quantity. Therefore, we have already been through the
        # normalization process and we simply want to pass the existing set of
        # arguments along to the target operator.
        if isinstance(args, Arguments):
            return self.context.apply(args)
        # Expand arguments to one index per dimension.
        normalized = self._normalize(args)
        nargs = len(normalized)
        ndims = len(self.dimensions)
        if nargs != ndims:
            raise IndexError(
                f"Number of indices ({nargs})"
                f" != number of dimensions ({ndims})"
            ) from None
        # Map each given index to its dimension.
        mapped = dict(zip(self.dimensions, normalized))
        axes = {k: v for k, v in mapped.items() if k != 'shell'}
        # Sort out the 'shell'/'radius' index. NOTE: If 'shell' is not in
        # self.dimensions, should we immediately return?
        if 'shell' in mapped:
            key, value = self._shell_or_radius(mapped)
            axes[key] = value
        # Implement this operation within the observing context. NOTE: The
        # design philosophy here assumes that each operator knows best how to
        # implement common array-manipulation tasks (e.g., interpolation). One
        # problem with that strategy is that the implementation step will raise
        # a `ValueError` before the operator has a chance to interpolate if a
        # physical axis value is not in the corresponding instance of
        # `Coordinates`.
        result = self.context.implement(**axes)
        # Check for unit consistency.
        if result.unit | self.unit:
            return result.withunit(self.unit)
        raise ObservationError(
            f"The unit of the observed array ({str(result.unit)!r})"
            f" is not consistent with the expected unit ({str(self.unit)!r})"
        ) from None

    def _normalize(self, args):
        """Observable-specific index normalization."""
        if self.ndim == 1 and quantity.ismeasurable(args):
            return numeric.index.expand(self.ndim, [args])
        return numeric.index.expand(self.ndim, args)

    def _shell_or_radius(self, axes: dict):
        """Determine whether the shell axis represents a radius."""
        shell = axes.get('shell', slice(None))
        if isinstance(shell, slice):
            return 'shell', shell
        s = quantity.measure(shell)
        if not quantity.isindexlike(s) and s.unit.quantity == 'length':
            return 'radius', physical.coordinates(s)
        return 'shell', physical.points(numpy.array(s))

    def __eq__(self, other):
        """Called for self == other."""
        return (
            isinstance(other, Quantity)
            and
            not self._unequal_metadata(other)
            and
            self.context == other.context
        )

    def __add__(self, other):
        """Called for self + other."""
        if not isinstance(other, Quantity):
            return NotImplemented
        if found := self._unequal_metadata(other):
            attrs = etc.join(found, quoted=True)
            raise ValueError(
                f"Cannot add quantities with unequal values of {attrs}"
            ) from None
        return self._apply(
            standard.add,
            other,
            self.unit,
            self.dimensions,
        )

    def __sub__(self, other):
        """Called for self - other."""
        if not isinstance(other, Quantity):
            return NotImplemented
        if found := self._unequal_metadata(other):
            attrs = etc.join(found, quoted=True)
            raise ValueError(
                f"Cannot subtract quantities with unequal values of {attrs}"
            ) from None
        return self._apply(
            standard.sub,
            other,
            self.unit,
            self.dimensions,
        )

    def __mul__(self, other):
        """Called for self * other."""
        if not isinstance(other, Quantity):
            return NotImplemented
        unit = self.unit * other.unit
        dimensions = self.dimensions | other.dimensions
        return self._apply(standard.mul, other, unit, dimensions)

    def __truediv__(self, other):
        """Called for self / other."""
        if not isinstance(other, Quantity):
            return NotImplemented
        unit = self.unit / other.unit
        dimensions = self.dimensions | other.dimensions
        return self._apply(standard.truediv, other, unit, dimensions)

    def __pow__(self, other):
        """Called for self ** other."""
        if not isinstance(other, numbers.Real):
            return NotImplemented
        unit = self.unit ** other
        return self._apply(
            builtins.pow,
            other,
            unit,
            self.dimensions,
        )

    def _unequal_metadata(self, other):
        """Return names of unequal metadata attributes."""
        if not isinstance(other, Quantity):
            raise TypeError(other)
        attrs = {'unit', 'dimensions'}
        unequal = [
            attr for attr in attrs
            if getattr(self, attr) != getattr(other, attr)
        ]
        return tuple(unequal)

    def _apply(self, operator, other, *attrs):
        """Apply a standard algebraic operator."""
        def __f(user: Arguments):
            return operator(self[user], other[user])
        context = Context(__f, self.context.axes, self.context.grid)
        return Quantity(context, *attrs)

    def withunit(self, new: metric.UnitLike | None, /):
        """Convert this quantity to a new unit."""
        if new is None or self.unit == new:
            return self
        if self.unit | new:
            unit = metric.unit(new)
            return Quantity(self.context, unit, self.dimensions)
        raise ValueError(
            f"The unit {str(new)!r} is inconsistent"
            f" with {str(self.unit)!r}"
        ) from None

    @property
    def context(self):
        """The relevant observing context."""
        return self._context

    @property
    def ndim(self):
        """The number of indexable dimensions."""
        if self._ndim is None:
            self._ndim = len(self.dimensions)
        return self._ndim

    @property
    def unit(self):
        """The metric unit of this quantity."""
        return self._unit

    @property
    def dimensions(self):
        """The indexable dimensions of this quantity."""
        return self._dimensions





import typing

from .. import base
from .. import metric
from . import _types
from . import _context
from ..typehelp import Self


I = typing.TypeVar('I', bound=_types.Object)


class Object(_types.Object[base.RealType]):
    """An object with real-valued data and an associated metric unit."""

    def __init__(
        self,
        data: base.RealType,
        context: _context.Context[I],
    ) -> None:
        super().__init__(data, context.unit)
        self._context = context

    def withunit(self: Self, new: metric.UnitLike, /) -> Self:
        """Convert this object to a new unit."""
        return self._context.convert(self, new)

    @property
    def context(self):
        """This object's initialization context."""
        return self._context


ObjectT = typing.TypeVar('ObjectT', bound=Object)
"""Type variable bounded by `~measured.Object`."""





import re

from . import _part


class Operator(_part.Part):
    """An operator in a symbolic expression."""

    def __init__(self, operation: str) -> None:
        self.operation = operation

    def __str__(self) -> str:
        """A simplified representation of this object."""
        return self.operation

    def __hash__(self) -> str:
        """Compute instance hash (e.g., for use as `dict` key)."""
        return hash(self.operation)

    def __eq__(self, other) -> bool:
        """True if two operators represent the same operation."""
        if isinstance(other, Operator):
            return other.operation == self.operation
        if isinstance(other, str):
            return other == self.operation
        return NotImplemented


class OperatorFactory(_part.Factory):
    """A factory that produces symbolic operators."""

    def __init__(
        self,
        multiply: str='*',
        divide: str='/',
    ) -> None:
        mul = fr'\{multiply}'
        div = fr'\{divide}'
        self.patterns = {
            'multiply': re.compile(
                fr'(?<!{div})(\s*{mul}\s*)(?!{div})'
            ),
            'divide': re.compile(
                fr'(?<!{mul})(\s*{div}\s*)(?!{mul})'
            ),
            'sqrt': re.compile(r'\s*sqrt\s*')
        }
        """Compiled regular expressions for symbolic operators."""

    def parse(self, string: str):
        """Extract an operator at the start of `string`, possible."""
        for key in self.patterns:
            if match := self.patterns[key].match(string):
                return _part.Match(
                    result=Operator(key),
                    context=match,
                )





"""
Abstract base classes corresponding to runtime protocols.
"""

import abc
import typing

import numpy

from .. import etc


A = typing.TypeVar('A')
B = typing.TypeVar('B')
R = typing.TypeVar('R')

BoolT = typing.Union[bool, typing.Sequence[bool]]


class Quantity(abc.ABC):
    """Abstract base class for all quantities.
    
    Similar in spirit to `numbers.Number`.
    """

    __slots__ = ()

    __hash__ = None


class Ordered(Quantity):
    """Abstract base class for ordered objects.

    A concrete implementation must define the following methods

    - `__lt__` (less than: `a < b`)
    - `__le__` (less than or equal: `a <= b`)
    - `__gt__` (greater than: `a > b`)
    - `__ge__` (greater than or equal: `a >= b`)

    See Also
    --------
    `~quantity.Ordered`
        The corresponding abstract protocol.
    """

    __slots__ = ()

    @abc.abstractmethod
    def __lt__(self, other) -> BoolT:
        """Called for self < other."""

    @abc.abstractmethod
    def __le__(self, other) -> BoolT:
        """Called for self <= other."""

    @abc.abstractmethod
    def __gt__(self, other) -> BoolT:
        """Called for self > other."""

    @abc.abstractmethod
    def __ge__(self, other) -> BoolT:
        """Called for self >= other."""


class Comparable(Ordered):
    """Abstract base class for comparable objects.

    A concrete implementation must define the following methods

    - `__eq__` (equal: `a == b`)
    - `__ne__` (not equal: `a != b`)
    - `__lt__` (inherited from `~Ordered`)
    - `__le__` (inherited from `~Ordered`)
    - `__gt__` (inherited from `~Ordered`)
    - `__ge__` (inherited from `~Ordered`)

    See Also
    --------
    `~quantity.Comparable`
        The corresponding abstract protocol.
    `~Ordered`
        Abstract base class for ordered objects.
    """

    __slots__ = ()

    @abc.abstractmethod
    def __eq__(self, other) -> BoolT:
        """Called for self == other."""

    @abc.abstractmethod
    def __ne__(self, other) -> BoolT:
        """Called for self != other."""


class Additive(Quantity):
    """Abstract base class for additive objects.

    A concrete implementation must define the following methods

    - `__add__`  (addition: `a + b`)
    - `__radd__` (addition with reflected operands: `b + a`)
    - `__sub__`  (subtraction: `a - b`)
    - `__rsub__` (subtraction with reflected operands: `b - a`)

    See Also
    --------
    `~quantity.Additive`
        The corresponding abstract protocol.
    """

    __slots__ = ()

    @abc.abstractmethod
    def __add__(self: A, other: B) -> typing.Union[A, B]:
        """Called for self + other."""

    @abc.abstractmethod
    def __radd__(self: A, other: B) -> typing.Union[A, B]:
        """Called for other + self."""

    @abc.abstractmethod
    def __sub__(self: A, other: B) -> typing.Union[A, B]:
        """Called for self - other."""

    @abc.abstractmethod
    def __rsub__(self: A, other: B) -> typing.Union[A, B]:
        """Called for other - self."""


class Multiplicative(Quantity):
    """Abstract base class for multiplicative objects.

    A concrete implementation must define the following methods

    - `__mul__`      (multiplication: `a * b`)
    - `__rmul__`     (multiplication with reflected operands: `b * a`)
    - `__truediv__`  (division: `a / b`)
    - `__rtruediv__` (division with reflected operands: `b / a`)

    See Also
    --------
    `~quantity.Multiplicative`
        The corresponding abstract protocol.
    """

    __slots__ = ()

    @abc.abstractmethod
    def __mul__(self: A, other: B) -> typing.Union[A, B]:
        """Called for self * other."""

    @abc.abstractmethod
    def __rmul__(self: A, other: B) -> typing.Union[A, B]:
        """Called for other * self."""

    @abc.abstractmethod
    def __truediv__(self: A, other: B) -> typing.Union[A, B]:
        """Called for self / other."""

    @abc.abstractmethod
    def __rtruediv__(self: A, other: B) -> typing.Union[A, B]:
        """Called for other / self."""


class Algebraic(Additive, Multiplicative):
    """Abstract base class for algebraic objects.
    
    A concrete implementation must define the following methods

    - `__add__`      (inherited from `~Additive`)
    - `__radd__`     (inherited from `~Additive`)
    - `__sub__`      (inherited from `~Additive`)
    - `__rsub__`     (inherited from `~Additive`)
    - `__mul__`      (inherited from `~Multiplicative`)
    - `__rmul__`     (inherited from `~Multiplicative`)
    - `__truediv__`  (inherited from `~Multiplicative`)
    - `__rtruediv__` (inherited from `~Multiplicative`)
    - `__pow__`      (exponentiation: `a ** b`)

    See Also
    --------
    `~quantity.Algebraic`
        The corresponding abstract protocol.
    `~Additive`
        Abstract base class for additive objects.
    `~Multiplicative`
        Abstract base class for multiplicative objects.
    """

    __slots__ = ()

    @abc.abstractmethod
    def __pow__(self: A, other: B) -> typing.Union[A, B]:
        """Called for self ** other."""


class Complex(Algebraic):
    """Abstract base class for complex-valued objects.

    A concrete implementation must define all methods required by `~Algebraic`,
    as well as the following methods

    - `__abs__` (absolute value: `abs(a)`)
    - `__pos__` (positive: `+a`)
    - `__neg__` (negative: `-a`)

    See Also
    --------
    `~quantity.Complex`
        The corresponding abstract protocol
    `~Algebraic`
        Abstract base class for algebraic objects.
    """

    __slots__ = ()

    @abc.abstractmethod
    def __abs__(self: A) -> A:
        """Called for abs(self)."""

    @abc.abstractmethod
    def __pos__(self: A) -> A:
        """Called for +self."""

    @abc.abstractmethod
    def __neg__(self: A) -> A:
        """Called for -self."""


class Real(Comparable, Complex):
    """Abstract base class for real-valued objects.

    This ABC is designed to encapsulate the behavior of objects that provide an
    interface to, or a container for, real-valued numerical data. It exists
    outside of the `numbers.Number` hierarchy in order to equally support
    single-valued and multi-valued objects.

    A concrete real-valued object must implement all methods required by the
    `~quantity.Real` protocol. Those include the following binary comparison
    operators

    - `__eq__` (inherited from `~Comparable`)
    - `__ne__` (inherited from `~Comparable`)
    - `__lt__` (inherited from `~Comparable`)
    - `__le__` (inherited from `~Comparable`)
    - `__gt__` (inherited from `~Comparable`)
    - `__ge__` (inherited from `~Comparable`)

    the following unary arithmetic operators

    - `__abs__` (inherited from `~Complex`)
    - `__pos__` (inherited from `~Complex`)
    - `__neg__` (inherited from `~Complex`)

    and the following binary arithmetic operators

    - `__add__`       (inherited from `~Algebraic`)
    - `__radd__`      (inherited from `~Algebraic`)
    - `__sub__`       (inherited from `~Algebraic`)
    - `__rsub__`      (inherited from `~Algebraic`)
    - `__mul__`       (inherited from `~Algebraic`)
    - `__rmul__`      (inherited from `~Algebraic`)
    - `__truediv__`   (inherited from `~Algebraic`)
    - `__rtruediv__`  (inherited from `~Algebraic`)
    - `__pow__`       (inherited from `~Algebraic`)
    - `__rpow__`      (exponentiation with reflected operands: `b ** a`)
    - `__floordiv__`  (floor division: `a // b`)
    - `__rfloordiv__` (floor division with reflected operands: `b // a`)
    - `__mod__`       (modular division: `a % b`)
    - `__rmod__`      (modular division with reflected operands: `b % a`)

    See https://docs.python.org/3/reference/datamodel.html for a description of
    each method and implementation suggestions.

    See Also
    --------
    `~quantity.Real`
        The corresponding abstract protocol.
    `~Comparable`
        Abstract base class for comparable objects.
    `~Complex`
        Abstract base class for complex-valued objects.
    """

    __slots__ = ()

    @abc.abstractmethod
    def __rpow__(self: A, other: B) -> typing.Union[A, B]:
        """Called for other ** self."""

    @abc.abstractmethod
    def __floordiv__(self: A, other: B) -> typing.Union[A, B]:
        """Called for self // other."""

    @abc.abstractmethod
    def __rfloordiv__(self: A, other: B) -> typing.Union[A, B]:
        """Called for other // self."""

    @abc.abstractmethod
    def __mod__(self: A, other: B) -> typing.Union[A, B]:
        """Called for self % other."""

    @abc.abstractmethod
    def __rmod__(self: A, other: B) -> typing.Union[A, B]:
        """Called for other % self."""


@etc.autostr
class Scalar(Real):
    """ABC for singular real-valued objects.
    
    A concrete scalar object must implement all methods required by the
    `~quantity.Scalar` protocol. In addition to those required by the
    `~quantity.Real` protocol, they include the following

    - `__int__`
    - `__floor__`
    - `__complex__`
    - `__round__`

    See Also
    --------
    `~quantity.Scalar`
        The corresponding abstract protocol.
    `~Real`
        Abstract base class for real-valued objects.
    """

    __slots__ = ()

    @abc.abstractmethod
    def __int__(self) -> int:
        """Called for int(self)."""

    @abc.abstractmethod
    def __float__(self) -> float:
        """Called for float(self)."""

    @abc.abstractmethod
    def __complex__(self) -> complex:
        """Called for complex(self)."""

    @abc.abstractmethod
    def __round__(self: A) -> A:
        """Called for round(self)."""


@etc.autostr
class Sequence(Quantity):
    """Abstract base class for sequence-like objects.

    A concrete implementation must default the following methods

    - `__contains__`
    - `__len__`
    - `__iter__`
    - `__getitem__`
    - `__array__`

    See Also
    --------
    `~quantity.Sequence`
        The corresponding abstract protocol.
    """

    __slots__ = ()

    @abc.abstractmethod
    def __contains__(self, v, /) -> bool:
        """Called for v in self."""

    @abc.abstractmethod
    def __len__(self) -> int:
        """Called for len(self)."""

    @abc.abstractmethod
    def __iter__(self) -> typing.Iterator:
        """Called for iter(self)."""

    @abc.abstractmethod
    def __getitem__(self, i: typing.SupportsIndex, /):
        """Called for self[i]."""

    @abc.abstractmethod
    def __array__(self, *args, **kwargs) -> numpy.ndarray:
        """Called for conversion to a `numpy.ndarray`."""


@etc.autostr
class Array(Sequence):
    """ABC for array-like objects.
    
    A concrete implementation must define all the methods and properties
    required by the `~quantity.Array` protocol. Those include the following
    methods (inherited from `~Sequence`)

    - `__iter__`
    - `__contains__`
    - `__len__`
    - `__getitem__`
    - `__array__`

    and the following properties

    - `size`
    - `ndim`
    - `shape`

    See Also
    --------
    `~quantity.Array`
        The corresponding abstract protocol.
    `~Sequence`
        Abstract base class for sequence-like objects.
    """

    __slots__ = ()

    @property
    @abc.abstractmethod
    def size(self) -> int:
        """The total number of values in this array."""

    @property
    @abc.abstractmethod
    def ndim(self) -> int:
        """The number of dimensions in this array."""

    @property
    @abc.abstractmethod
    def shape(self) -> typing.Tuple[int]:
        """The length of each dimension in this array."""




import abc
import typing

from .. import base
from .. import measured
from .. import metric
from .. import numeric
from .. import quantity


class Quantity(measured.Object[base.RealType], base.abc.Real):
    """Abstract base class for all physical objects."""

QuantityType = typing.TypeVar('QuantityType', bound=Quantity)


T = typing.TypeVar('T')


class Protocol(measured.Interface, base.Real, typing.Protocol[T]):
    """Abstract protocol class for physical objects.
    
    A physical object is a real-valued measurable object. In addition to all
    abstract properties and methods inherited from from `~measured.Protocol`, a
    physical object must implement the real-valued protocol (defined by
    `~base.Real`) and must implement the `__measure__` method.
    """

    @abc.abstractmethod
    def __measure__(self) -> quantity.Measurement:
        """Called for `~quantity.measure(self)`."""
        raise NotImplementedError


@typing.runtime_checkable
class Type(Protocol, typing.Protocol[T]):
    """Structural protocol clas for physical objects."""

    def __measure__(self) -> quantity.Measurement: ...


@typing.runtime_checkable
class ArrayType(numeric.Type, metric.Type, typing.Protocol):
    """Structural protocol class for physical arrays.

    Objects that define the following properties will pass instance checks
    against this protocol

    - `data`
    - `unit`
    - `isunitless`
    - `dimensions`

    See Also
    --------
    `~numeric.Type`
        The protocol class that requires `data`.
    `~metric.Type`
        The protocol class that requires `unit` and `isunitless`.
    """

    @property
    def dimensions(self) -> numeric.Dimensions: ...


